/* Copyright (C) 2007 Christian Dywan <christian at twotoasts dot de>

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA. */

#include "sokoke.h"

void sokoke_widget_set_visible(GtkWidget* widget, gboolean bVisibility)
{
    // Show or hide the widget
    if(bVisibility) gtk_widget_show(widget); else gtk_widget_hide(widget);
}

void sokoke_container_show_children(GtkContainer* container)
{
    // Show every child but not the container itself
    // Warnung: Übergabe des Arguments 2 von »gtk_container_foreach« von
    // inkompatiblem Zeigertyp ?
    gtk_container_foreach(container, (GtkCallback)(gtk_widget_show_all), NULL);
}

GtkWidget* sokoke_hig_frame_new(const gchar* sLabel)
{
    // Create a frame with no actual frame but a bold label (and indentation)
    // TODO: We lack proper indentation, can we solve this w/o an actual widget?
    GtkWidget* frame = gtk_frame_new(NULL);
    gchar* sLabelBold = g_strdup_printf("<b>%s</b>", sLabel);
    GtkWidget* label = gtk_label_new(NULL);
    gtk_label_set_markup(GTK_LABEL(label), sLabelBold);
    g_free(sLabelBold);
    gtk_frame_set_label_widget(GTK_FRAME(frame), label);
    gtk_frame_set_shadow_type(GTK_FRAME(frame), GTK_SHADOW_NONE);
    return frame;
}

gchar* sokoke_key_file_get_string_default(GKeyFile* key_file
 , const gchar* group_name, const gchar* key, const gchar* def, GError* *error)
{
    gchar* value = g_key_file_get_string(key_file, group_name, key, error);
    if(value == NULL) return g_strdup(def);
    return value;
}

gint sokoke_key_file_get_integer_default(GKeyFile* key_file
 , const gchar* group_name, const gchar* key, const gint def, GError* *error)
{
    gint value = g_key_file_get_integer(key_file, group_name, key, error);
    if(error == NULL) return def;
    return value;
}

gboolean sokoke_key_file_save_to_file(GKeyFile* key_file
 , const gchar* file, GError* *error)
{
    gchar* data = g_key_file_to_data(key_file, NULL, error);
    //if(error != NULL) return FALSE;
    FILE *fp;
    if((fp = fopen(file, "w")) == NULL) return FALSE;
    fprintf(fp, data);
    fclose(fp);
    g_free(data);
    return TRUE;
}

void sokoke_widget_get_text_size(GtkWidget* widget, const gchar* sText
 , gint* w, gint* h)
{
    PangoLayout* layout = gtk_widget_create_pango_layout(widget, sText);
    pango_layout_get_pixel_size(layout, w, h);
    g_object_unref(layout);
}

#define QUIT_LOOP(ri) \
 if(g_main_loop_is_running(ri->loop)) \
  g_main_loop_quit (ri->loop)

static void sokoke_unmap_h(GtkDialog* dialog, sokoke_ri* ri)
{
    QUIT_LOOP(ri);
}

static void sokoke_response_h(GtkDialog*dialog, gint response_id, sokoke_ri* ri)
{
    ri->response_id = response_id;
    QUIT_LOOP(ri);
}

static gint sokoke_delete_h(GtkDialog* dialog, GdkEventAny* event, sokoke_ri* ri)
{
    QUIT_LOOP(ri);
    return TRUE;
}

static void sokoke_destroy_h(GtkDialog* dialog, sokoke_ri* ri)
{
    // QUIT_LOOP will be called in sokoke_unmap_h
    ri->destroyed = TRUE;
}

gint sokoke_dialog_run_modeless(GtkDialog* dialog)
{
    // Based on gtk_dialog_run from gtk/gtkdialog.c from the GTK+ project.
    // Copyright (C) 1995-1997 Peter Mattis, Spencer Kimball and Josh MacDonald

    // This will display a modeless dialog and return the response

    sokoke_ri ri = { NULL, GTK_RESPONSE_NONE, NULL, FALSE };

    g_return_val_if_fail(GTK_IS_DIALOG(dialog), -1);
    g_object_ref(dialog);

    if(!GTK_WIDGET_VISIBLE(dialog)) gtk_widget_show(GTK_WIDGET(dialog));

    gulong response_h = g_signal_connect(dialog, "response"
     , GTK_SIGNAL_FUNC(sokoke_response_h), &ri);
    gulong unmap_h = g_signal_connect(dialog, "unmap"
     , GTK_SIGNAL_FUNC(sokoke_unmap_h), &ri);
    gulong delete_h = g_signal_connect(dialog, "delete-event"
     , GTK_SIGNAL_FUNC(sokoke_destroy_h), &ri);
    gulong destroy_h = g_signal_connect(dialog, "destroy"
     , GTK_SIGNAL_FUNC(sokoke_destroy_h), &ri);

    ri.loop = g_main_loop_new (NULL, FALSE);
    GDK_THREADS_LEAVE();
    g_main_loop_run(ri.loop);
    GDK_THREADS_ENTER();
    g_main_loop_unref(ri.loop);
    ri.loop = NULL;

    if(!ri.destroyed)
    {
      g_signal_handler_disconnect(dialog, response_h);
      g_signal_handler_disconnect(dialog, unmap_h);
      g_signal_handler_disconnect(dialog, delete_h);
      g_signal_handler_disconnect(dialog, destroy_h);
    }

    g_object_unref (dialog);
    return ri.response_id;
}

#undef QUIT_LOOP

void sokoke_widget_update_screen(GtkWidget* widget
 , GdkScreen* old_screen, gpointer userdata)
{
    g_print("%i\n", (gboolean)g_object_get_data(G_OBJECT(widget), "sokoke_bAlpha"));
    GdkScreen* screen = gtk_widget_get_screen(widget);
    GdkColormap* colormap;
    #if GTK_CHECK_VERSION(2, 8, 0)
    colormap = gdk_screen_get_rgba_colormap(screen);
    if(!colormap)
    {
    #endif
     colormap = gdk_screen_get_rgb_colormap(screen);
     g_object_set_data(G_OBJECT(widget), "sokoke_bAlpha", (gpointer)FALSE);
    #if GTK_CHECK_VERSION(2, 8, 0)
    }
    else
     g_object_set_data(G_OBJECT(widget), "sokoke_bAlpha", (gpointer)TRUE);
    #endif
    gtk_widget_set_colormap(widget, colormap);
}

gboolean sokoke_on_transparent_widget_expose(GtkWidget* widget
 , GdkEventExpose* event, gpointer userdata)
{
    #if GTK_CHECK_VERSION(2, 88, 0)
    g_print("%i\n", (gboolean)g_object_get_data(G_OBJECT(widget), "sokoke_bAlpha"));
    // TODO: Fix xorg errors
    // TODO: Remove circle code
    // TODO: Fallback to pseudo transparency even for GTK+ 2.8
    cairo_t *cr = gdk_cairo_create(widget->window);
    if((gboolean)g_object_get_data(G_OBJECT(widget), "sokoke_bAlpha"))
     cairo_set_source_rgba(cr, 1.0, 1.0, 1.0
      , (gfloat)g_object_get_data(G_OBJECT(widget), "sokoke_fAlpha"));
    else
     cairo_set_source_rgb(cr, 1.0, 1.0, 1.0);

    // We have real transparency:
    // Make the widget's background transparent
    cairo_set_operator(cr, CAIRO_OPERATOR_SOURCE);
    cairo_paint(cr);

    int width, height;
    gtk_window_get_size(GTK_WINDOW(widget), &width, &height);

    cairo_set_source_rgba(cr, 1, 0.2, 0.2, 0.6);
    cairo_arc(cr, width / 2, height / 2, (width < height ? width : height) / 2 - 8 , 0, 2 * 3.14);
    cairo_fill(cr);
    cairo_stroke(cr);

    cairo_destroy(cr);
    #else
    // We have pseudo transparency:
    // Draw the root window's content on the widget's background
    GdkPixbuf* bg;
    int x, y, w, h;

    gdk_window_get_root_origin(widget->window, &x, &y);
    gdk_drawable_get_size(widget->window, &w, &h);
    bg = gdk_pixbuf_get_from_drawable(NULL, gtk_widget_get_root_window(widget)
     , NULL, x, y, 0, 0, w, h);
    gdk_draw_pixbuf(widget->window, NULL, bg, 0, 0, 0, 0, w, h
     , GDK_RGB_DITHER_NONE, 0, 0);
    #endif
    return FALSE;
}


void sokoke_widget_set_transparency(GtkWidget* widget, gfloat fAlpha)
{
    gtk_widget_set_app_paintable(widget, fAlpha > 0);
    g_object_set_data(G_OBJECT(widget), "sokoke_fAlpha", (gpointer)&fAlpha);
    if(fAlpha)
    {
     g_signal_connect(G_OBJECT(widget), "expose-event"
      , G_CALLBACK(sokoke_on_transparent_widget_expose), NULL);
     g_signal_connect(G_OBJECT(widget), "screen-changed"
      , G_CALLBACK(sokoke_widget_update_screen), NULL);
     sokoke_widget_update_screen(widget, NULL, NULL);
    }
    else
    {
     // TODO: disconnect signals
    }
}

void sokoke_menu_item_set_accel(GtkMenuItem* menuitem, const gchar* sPath
 , const gchar* sKey, GdkModifierType accel_mods)
{
    if(strlen(sPath))
    {
     gchar* path = g_strconcat("<", g_get_prgname(), ">/", sPath, NULL);
     gtk_menu_item_set_accel_path(GTK_MENU_ITEM(menuitem), path);
     guint keyVal = sKey ? gdk_keyval_from_name(sKey) : 0;
     gtk_accel_map_add_entry(path, keyVal, accel_mods);
     g_free(path);
    }
}

gboolean sokoke_entry_can_undo(GtkEntry* entry)
{
    // TODO: Can we undo the last input?
    return FALSE;
}

gboolean sokoke_entry_can_redo(GtkEntry* entry)
{
    // TODO: Can we redo the last input?
    return FALSE;
}

void sokoke_entry_undo(GtkEntry* entry)
{
    // TODO: Implement undo
}

void sokoke_entry_redo(GtkEntry* entry)
{
    // TODO: Implement redo
}

gboolean sokoke_on_undo_entry_key_down(GtkEntry* widget, GdkEventKey* event
 , gpointer userdata)
{
    switch(event->keyval)
    {
     case GDK_Undo:
      sokoke_entry_undo(widget);
      return FALSE;
     case GDK_Redo:
      sokoke_entry_redo(widget);
      return FALSE;
     default:
      return FALSE;
    }
}

void sokoke_on_undo_entry_populate_popup(GtkEntry* entry, GtkMenu* menu
 , gpointer userdata)
{
    // Enhance the entry's menu with undo and redo items.
    GtkWidget* menuitem = gtk_separator_menu_item_new();
    gtk_menu_shell_prepend((GtkMenuShell*)menu, menuitem);
    gtk_widget_show(menuitem);
    menuitem = gtk_image_menu_item_new_from_stock(GTK_STOCK_REDO, NULL);
    g_signal_connect(menuitem, "activate", GTK_SIGNAL_FUNC(sokoke_entry_redo), userdata);
    gtk_menu_shell_prepend((GtkMenuShell*)menu, menuitem);
    gtk_widget_show(menuitem);
    menuitem = gtk_image_menu_item_new_from_stock(GTK_STOCK_UNDO, NULL);
    g_signal_connect(menuitem, "activate", GTK_SIGNAL_FUNC(sokoke_entry_undo), userdata);
    gtk_menu_shell_prepend((GtkMenuShell*)menu, menuitem);
    gtk_widget_show(menuitem);
}

gboolean sokoke_entry_get_can_undo(GtkEntry* entry)
{
    // TODO: Is this entry undo enabled?
    return FALSE;
}

void sokoke_entry_set_can_undo(GtkEntry* entry, gboolean bCanUndo)
{
    if(bCanUndo)
    {
     g_signal_connect(G_OBJECT(entry), "key-press-event"
      , G_CALLBACK(sokoke_on_undo_entry_key_down), NULL);
     g_signal_connect(G_OBJECT(entry), "populate-popup"
      , G_CALLBACK(sokoke_on_undo_entry_populate_popup), NULL);
    }
    else
    {
     ; // TODO: disconnect signal
    }
}
